/*
 */

#include <linux/module.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include "lcd.h"

/* the BIOS manuals say there can be up to 4 lpt devices
 * but I have not seen a board where the 4th address is listed
 * if you have different hardware change the table below
 * please let me know if you have different equipment
 * if you have more than 3 printers, remember to increase LP_NO
 */
struct lcd_struct lcd_table[] = {
	{ 0x3bc, 0, 0 },
	{ 0x378, 0, 0 },
	{ 0x278, 0, 0 },
};
#define LCD_NO 3

#define LCD_MAJOR   60   /* for experimental use */

#define LCD_WAIT_TIME 1000


static int lcd_write_ctrl( int minor, int data );
static int lcd_write_data( int minor, int data );
static int lcd_read_ctrl( int minor );
static int lcd_read_data( int minor );
static int lcd_reset_pwsw( int minor );
static int lcd_read_pwsw( int minor );
static int lcd_wait_ready( int minor );

static int lcd_disp_move_pos( int minor, int x, int y );
static int lcd_disp_scroll_up( int minor );
static int lcd_disp_putc( int minor, int d );
static int lcd_disp_puts( int minor, char *str );

static char *initial_string = "\x0cLCD driver\nReady.";

static int
lcd_reset(int minor)
{
        int i;

        lcd_reset_pwsw( minor );
	if( lcd_read_pwsw( minor ) ){
	    printk(KERN_INFO "lcd%d: pwsw ctrl not ready\n", minor );
	    return -1;
	}

/* reset LCD module */
	lcd_write_ctrl( minor, 0x30 );
	for( i=0; i<5; ++i ){
	    udelay( 1000 );
	}

	lcd_write_ctrl( minor, 0x30 );
	udelay( 1000 );

	lcd_write_ctrl( minor, 0x30 );
	udelay( 1000 );

/* selsect 8bit bus mode, 2lines mode, 5x8dots fonts */
	lcd_write_ctrl( minor, 0x38 );
/* Dispkay on, Cursor on, Cursor Blinking off */
	lcd_write_ctrl( minor, 0xe );
/* Cursor move direction is increment */
	lcd_write_ctrl( minor, 0x6 );

	lcd_disp_puts( minor, initial_string );

	return 0;
}


static int
lcd_lseek(struct inode * inode, struct file * file,
		    off_t offset, int origin)
{
	return -ESPIPE;
}

static int
lcd_open(struct inode * inode, struct file * file)
{
        static int first_use_flag = 1;

	unsigned int minor = MINOR(inode->i_rdev);

	if (minor >= LCD_NO)
		return -ENXIO;
	if ((LCD_F(minor) & LCD_EXIST) == 0)
		return -ENXIO;
	if (LCD_F(minor) & LCD_BUSY)
		return -EBUSY;

	MOD_INC_USE_COUNT;

	LCD_F(minor) |= LCD_BUSY;

	if( first_use_flag ){
	    first_use_flag = 0;
	    lcd_disp_putc( minor, '\x0c' );
	}

#ifdef LCD_DEBUG
	printk("lcd%d open\n", minor );
#endif

	return 0;
}

static void
lcd_release(struct inode * inode, struct file * file)
{
	unsigned int minor = MINOR(inode->i_rdev);

	LCD_F(minor) &= ~LCD_BUSY;
	MOD_DEC_USE_COUNT;

#ifdef LCD_DEBUG
	printk("lcd%d release\n", minor );
#endif
}

#define KBUFF_SIZE 32

static int
lcd_write( struct inode *inode, struct file *file, const char *buf, int count )
{
        static char kbuff[KBUFF_SIZE];
	unsigned int minor = MINOR(inode->i_rdev);
	int  retval = 0;
	int  size, i;

#ifdef LCD_DEBUG
	printk("lcd%d write, count = %d\n", minor, count );
#endif

	retval = verify_area( VERIFY_READ, buf, count );
	if (retval){
	    return retval;
	}
	size = (count > KBUFF_SIZE)?(KBUFF_SIZE):(count);
	memcpy_fromfs( kbuff, buf, size );

	for( i=0; i<size; ++i){
	    lcd_disp_putc( minor, (int)kbuff[i] );
	}

	return size;
}


static int
lcd_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	unsigned int minor = MINOR(inode->i_rdev);
	int retval = 0;
	int parm;

#ifdef LCD_DEBUG
	printk("lcd%d ioctl, cmd: 0x%x\n",
	       minor, cmd );
#endif

	retval = verify_area(VERIFY_READ, (void *) arg, sizeof(int));
	if (retval){
	    return retval;
	}
	memcpy_fromfs( &parm, (void *)arg, sizeof(int) );

#ifdef LCD_DEBUG
	printk("lcd%d ioctl, parm: 0x%x\n", minor, parm );
#endif

	if (minor >= LCD_NO)
		return -ENODEV;
	if ((LCD_F(minor) & LCD_EXIST) == 0)
		return -ENODEV;
	switch ( cmd ) {
		case LCD_WRITE_DATA:
		        lcd_write_data( minor, parm );
		        break;
	        case LCD_READ_DATA:
			parm = lcd_read_data( minor );
		        break;
	        case LCD_WRITE_CTRL:
		        lcd_write_ctrl( minor, parm );
			break;
	        case LCD_READ_CTRL:
			parm = lcd_read_ctrl( minor);
		        break;
	        case LCD_READ_PWSW:
			parm = lcd_read_pwsw( minor );
		        break;
	        case LCD_RESET_PWSW:
			lcd_reset_pwsw( minor );
		        break;
		default:
			retval = -EINVAL;
	}

	if( retval == 0 ){
	    retval = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));
	    if (retval){
		return retval;
	    }
	    memcpy_tofs( (void *)arg, &parm, sizeof(int) );
	}

	return retval;
}

/*****************************/

static int
lcd_disp_move_pos( int minor, int x, int y )
{
    if( ( x<0 || x>15 ) || ( y<0 || y>1 ) ){
	return -1;
    }
    lcd_write_ctrl( minor, 0x80+x+y*0x40 );

    return 0;
}

static int
lcd_disp_scroll_up( int minor )
{
    int  x, d;

    for( x = 0; x < 16; ++x ){
	lcd_disp_move_pos( minor, x, 1 );
	d = lcd_read_data( minor );
	lcd_disp_move_pos( minor, x, 0 );
	lcd_write_data( minor, d );
	lcd_disp_move_pos( minor, x, 1 );
	lcd_write_data( minor, 0x20 );
    }
    lcd_disp_move_pos( minor, 0, 1 );

    return 0;
}

static int
lcd_disp_putc( int minor, int d )
{
    int x, y, ad;

    while( (ad = lcd_read_ctrl( minor )) & 0x80 );

    x = ad % 0x40;
    y = ad / 0x40;

    switch( d ){
    case 0x0a:  /* ^J */
	if( y >= 1 ){
	    lcd_disp_scroll_up( minor );
	}
	lcd_disp_move_pos( minor, 0, 1 );
	break;
    case 0x0d: /* ^M */
	lcd_disp_move_pos( minor, 0, y );
	break;
    case 0x0c: /* ^L */
	lcd_write_ctrl( minor, 0x1 );  /* clear display */
	lcd_write_ctrl( minor, 0x2 );  /* go to home */
	break;
    case 0x08: /* ^H */
	if( x > 0 ){
	    --x;
	    lcd_disp_move_pos( minor, x, y );
	}
	break;
    default:
	if( x >= 16 ){
	    x = 0;
	    ++y;
	    if( y > 1 ){
		lcd_disp_scroll_up( minor );
		--y;
	    }
	    lcd_disp_move_pos( minor, x, y );
	}
	lcd_write_data( minor, d );
	break;
    }

    return 0;
}

static int
lcd_disp_puts( int minor, char *str )
{
    int d;

    while( (d = *str++ ) ){
	lcd_disp_putc( minor, d );
    }

    return 0;
}

/*****************************/


static int
lcd_write_ctrl( int minor, int data )
{
	if( lcd_wait_ready( minor) < 0 ){
	        return -1;
	}

        outb_p( 0x0a, LCD_C(minor) );
	udelay(1);
        outb_p( data, LCD_B(minor) );
	udelay(1);
        outb_p( 0x02, LCD_C(minor) );
	udelay(1);
        outb_p( 0x0a, LCD_C(minor) );
	udelay(1);

	return 0;
}


static int
lcd_write_data( int minor, int data )
{
	if( lcd_wait_ready( minor) < 0 ){
	        return -1;
	}

        outb_p( 0x0e, LCD_C(minor) );
	udelay(1);
        outb_p( data, LCD_B(minor) );
	udelay(1);
        outb_p( 0x06, LCD_C(minor) );
	udelay(1);
        outb_p( 0x0e, LCD_C(minor) );
	udelay(1);

	return 0;
}

static int
lcd_read_ctrl( int minor )
{
        int data;

        outb_p( 0x0e8, LCD_C(minor) );
	udelay(1);
        outb_p( 0xe0, LCD_C(minor) );
	udelay(1);
        data = inb_p( LCD_B(minor) );
        outb_p( 0xe8, LCD_C(minor) );
	udelay(1);

	return ( data & 0xff );
}


static int
lcd_read_data( int minor )
{
        int data;

	if( lcd_wait_ready( minor) < 0 ){
	        return -1;
	}

        outb_p( 0x0ec, LCD_C(minor) );
	udelay(1);
        outb_p( 0xe4, LCD_C(minor) );
	udelay(1);
        data = inb_p( LCD_B(minor) );
        outb_p( 0xec, LCD_C(minor) );
	udelay(1);

	return (data & 0xff);
}


static int
lcd_reset_pwsw( int minor )
{
        outb_p( 0x0b, LCD_C(minor) );
	udelay(1);
        outb_p( 0x0a, LCD_C(minor) );
	udelay(1);

	return 0;
}


static int
lcd_read_pwsw( int minor )
{
        int data;

	data = inb_p( LCD_S(minor) );

	if( data & 0x40 ){
	        return 1;
	}else{
	        return 0;
	}
}


static int
lcd_wait_ready( int minor )
{
        int i;

	for( i=0; i < LCD_WAIT_TIME; ++i ){
	        if( ( lcd_read_ctrl( minor ) & 0x80 ) == 0 ){
		        return 0;
		}
	}

	return -1;
}


static struct file_operations lcd_fops = {
	lcd_lseek,
	NULL,		/* lcd_read */
	lcd_write,      /* lcd_write */
	NULL,		/* lcd_readdir */
	NULL,		/* lcd_select */
	lcd_ioctl,
	NULL,		/* lcd_mmap */
	lcd_open,
	lcd_release
};

static int
lcd_probe(int offset)
{
	int base, size;
	unsigned int testvalue;

	base = LCD_B(offset);
#ifdef LCD_DEBUG
	printk(KERN_INFO "base =  0x%04x\n", base);
#endif
	if (base == 0){
		return -1;		/* de-configured by command line */
	}
#ifdef LCD_DEBUG
	printk(KERN_INFO "irq  =  0x%04x\n", LCD_IRQ(offset) );
#endif
	if (LCD_IRQ(offset) > 15){
		return -1;		/* bogus interrupt value */
	}
	size = (base == 0x3bc)? 3 : 8;
#ifdef LCD_DEBUG
	printk(KERN_INFO "size =  0x%04x\n", size );
#endif
	if (check_region(base, size) < 0){
	        printk(KERN_INFO "check_region failed.\n" );
		return -1;
	}
	/* write to port & read back to check */
	outb_p( 0x0a, base +2);
	udelay(LCD_DELAY);
	outb_p(LCD_DUMMY, base);
	udelay(LCD_DELAY);
	testvalue = inb_p(base);
	if (testvalue == LCD_DUMMY) {
		LCD_F(offset) |= LCD_EXIST;
		lcd_reset(offset);
		printk(KERN_INFO "lcd%d at 0x%04x, ", offset, base);
		request_region(base, size, "lcd");
		if (LCD_IRQ(offset))
			printk("(irq = %d)\n", LCD_IRQ(offset));
		else
			printk("(polling)\n");
		return 1;
	} else {
	        printk(KERN_INFO "dummy write failed.\n" );
		return 0;
	}
}

/* Command line parameters:

   When the lp driver is built in to the kernel, you may use the
   LILO/LOADLIN command line to set the port addresses and interrupts
   that the driver will use.

   Syntax:	lp=port0[,irq0[,port1[,irq1[,port2[,irq2]]]]]

   For example:   lp=0x378,0   or   lp=0x278,5,0x378,7

   Note that if this feature is used, you must specify *all* the ports
   you want considered, there are no defaults.  You can disable a
   built-in driver with lp=0 .

*/

void
lcd_setup(char *str, int *ints)
{	
        LCD_B(0)   = ((ints[0] > 0) ? ints[1] : 0 );
        LCD_IRQ(0) = ((ints[0] > 1) ? ints[2] : 0 );
        LCD_B(1)   = ((ints[0] > 2) ? ints[3] : 0 );
        LCD_IRQ(1) = ((ints[0] > 3) ? ints[4] : 0 );
        LCD_B(2)   = ((ints[0] > 4) ? ints[5] : 0 );
        LCD_IRQ(2) = ((ints[0] > 5) ? ints[6] : 0 );
}

#ifdef MODULE
static int io[] = {0, 0, 0};
static int irq[] = {0, 0, 0};

#define lcd_init init_module
#endif

int
lcd_init(void)
{
	int offset = 0;
	int count = 0;
#ifdef MODULE
	int failed = 0;
#endif

	if (register_chrdev( LCD_MAJOR, "lcd", &lcd_fops )) {
		printk("lcd: unable to get major %d\n", LCD_MAJOR);
		return -EIO;
	}
#ifdef MODULE
	/* When user feeds parameters, use them */
	for (offset=0; offset < LCD_NO; offset++) {
		int specified=0;

		if (io[offset] != 0) {
			LCD_B(offset) = io[offset];
			specified++;
		}
		if (irq[offset] != 0) {
			LCD_IRQ(offset) = irq[offset];
			specified++;
		}
		if (specified) {
			if (lcd_probe(offset) <= 0) {
				printk(KERN_INFO "lcd%d: Not found\n", offset);
				failed++;
			} else {
				count++;
			}
		}
	}
	/* Successful specified devices increase count
	 * Unsuccessful specified devices increase failed
	 */
	if (count)
		return 0;
	if (failed) {
		printk(KERN_INFO "lcd: No override devices found.\n");
		unregister_chrdev(LCD_MAJOR,"lcd");
		return -EIO;
	}
	/* Only get here if there were no specified devices. To continue 
	 * would be silly since the above code has scribbled all over the
	 * probe list.
	 */
#endif
	/* take on all known port values */
	for (offset = 0; offset < LCD_NO; offset++) {
		int ret = lcd_probe(offset);
		if (ret < 0){
			continue;
		}
		count += ret;
	}
	if (count == 0){
		printk("lcd: Driver configured but no interfaces found.\n");
	}

	return 0;
}

#ifdef MODULE
void
cleanup_module(void)
{
	int offset;

	unregister_chrdev(LCD_MAJOR,"lcd");
	for (offset = 0; offset < LCD_NO; offset++) {
		int base, size;
		base = LCD_B(offset);
		size = (base == 0x3bc)? 3 : 8;
		if (LCD_F(offset) & LCD_EXIST){
			release_region( base, size);
			printk(KERN_INFO "lcd%d at 0x%04x released.\n", offset, base);
		}
	}
}
#endif




